"""..."""

from dataclasses import asdict

from aligo.core import *
from aligo.core.Config import *
from aligo.request import *
from aligo.types import *


class Update(BaseAligo):
    """..."""

    def update_file(self, body: UpdateFileRequest) -> BaseFile:
        """
        Update file.
        :param body: [UpdateFileRequest]
        :return: [BaseFile]

        :Example:
        >>> from aligo import Aligo
        >>> ali = Aligo()
        >>> new_file = ali.update_file(UpdateFileRequest(file_id='file_id', name='new_name'))
        >>> print(new_file.name)
        """
        response = self._post(V3_FILE_UPDATE, body=body)
        return self._result(response, BaseFile)

    def _core_rename_file(self, body: RenameFileRequest) -> BaseFile:
        """..."""
        return self.update_file(UpdateFileRequest(**asdict(body)))



"""..."""
from dataclasses import field, dataclass

from aligo.types import DataClass


@dataclass
class SearchFileRequest(DataClass):
    """..."""
    query: str
    drive_id: str = None
    limit: int = field(default=100, repr=False)
    image_thumbnail_process: str = field(default='image/resize,w_160/format,jpeg', repr=False)
    image_url_process: str = field(default='image/resize,w_1920/format,jpeg', repr=False)
    marker: str = field(default=None, repr=False)
    order_by: str = field(default=None, repr=False)
    url_expire_sec: int = field(default=14400, repr=False)
    video_thumbnail_process: str = field(default='video/snapshot,t_0,f_jpg,ar_auto,w_800', repr=False)



"""兑换福利码响应"""
from dataclasses import dataclass
from typing import Dict

from aligo.types import DataClass


@dataclass
class Result:
    """..."""
    message: str = None


@dataclass
class RewardSpaceResponse(DataClass):
    """数据结构, 类型 可能不准确"""

    success: bool = None
    code: str = None
    message: str = None
    totalCount: int = None
    nextToken: str = None
    maxResults: Dict = None
    result: Result = None
    arguments: str = None

    def __post_init__(self):
        if self.result:
            self.result = Result(**self.result)



"""..."""
from dataclasses import dataclass
from typing import List

from .AudioMeta import AudioMeta
from .AudioMusicMeta import AudioMusicMeta
from .AudioTranscodeTemplate import AudioTranscodeTemplate
from .DataClass import DataClass
from .VideoPreviewSprite import VideoPreviewSprite
from .VideoTranscodeTemplate import VideoTranscodeTemplate


@dataclass
class VideoPreview(DataClass):
    """..."""
    video_format: str = None
    audio_format: str = None
    duration: str = None
    audio_sample_rate: str = None
    audio_channels: int = None
    audio_template_list: List[AudioTranscodeTemplate] = None
    audio_meta: AudioMeta = None
    audio_music_meta: AudioMusicMeta = None
    bitrate: str = None
    frame_rate: str = None
    height: int = None
    sprite_info: VideoPreviewSprite = None
    template_list: List[VideoTranscodeTemplate] = None
    thumbnail: str = None
    width: int = None



"""..."""
import os
import re
from typing import Iterator, List

import requests
from tqdm import tqdm

from aligo.core import *
from aligo.core.Config import *
from aligo.request import *
from aligo.response import *
from aligo.types import *


class Download(BaseAligo):
    """..."""
    _DOWNLOAD_CHUNK_SIZE = 8388608 * 10  # 8M * 10

    def _core_get_download_url(self, body: GetDownloadUrlRequest) -> GetDownloadUrlResponse:
        """..."""
        response = self._post(V2_FILE_GET_DOWNLOAD_URL, body=body)
        return self._result(response, GetDownloadUrlResponse)

    def _core_batch_download_url(self, body: BatchDownloadUrlRequest) -> Iterator[BatchDownloadUrlResponse]:
        """..."""
        if body.drive_id is None:
            body.drive_id = self.default_drive_id

        yield from self.batch_request(BatchRequest(
            requests=[BatchSubRequest(
                id=file_id,
                url='/file/get_download_url',
                body=GetDownloadUrlRequest(
                    drive_id=body.drive_id, file_id=file_id
                )
            ) for file_id in body.file_id_list]
        ), GetDownloadUrlResponse)

    @staticmethod
    def _del_special_symbol(s: str) -> str:
        """删除Windows文件名中不允许的字符"""
        return re.sub(r'[\\/:*?"<>|]', '_', s)

    def _core_download_file(self, file_path: str, url: str) -> str:
        """下载文件

        :param file_path: 下载到哪里, 比如: 123.jpg, ./123.jpg, D:\123.jpg
        :param url: 下载地址
        :return: 下载完成保存文件的本地路径
        """
        file_dir, file_name = os.path.split(os.path.abspath(file_path))
        file_name = self._del_special_symbol(file_name)
        file_path = os.path.join(file_dir, file_name)

        # url 为空判断
        if not url:
            self._auth.log.error(f'文件 {file_path} 下载失败: 获取的url为空, 文件可能被屏蔽导致无法下载')
            return file_path

        self._auth.log.info(f'开始下载文件 {file_path}')

        if os.path.exists(file_path) and not os.path.exists(f'{file_path}.aria2'):
            self._auth.log.warning(f'文件已存在,跳过下载 {file_path}')
            return file_path

        if self._has_aria2c:
            cmd = ' '.join([
                f'aria2c "{url}"',
                f'--referer=https://www.aliyundrive.com/',
                f'--console-log-level=warn',
                f'--download-result=hide',
                f'--auto-file-renaming=false',
                f'-d "{file_dir}"',
                f'-o "{file_name}"',
            ])
            # print(cmd)
            os.system(cmd)
            return file_path

        # 递归创建目录
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        tmp_file = file_path + '.ali'

        tmp_size = 0
        if os.path.exists(tmp_file):
            tmp_size = os.path.getsize(tmp_file)

        try:
            progress_bar = None
            with requests.get(url, headers={
                'referer': 'https://www.aliyundrive.com/',
                'Range': f'bytes={tmp_size}-'
            }, stream=True) as resp:
                llen = int(resp.headers.get('content-length', 0))
                if resp.headers.get('Accept-Ranges', None) != 'bytes':
                    raise f'此链接不支持断点续传 {resp.url}'
                progress_bar = tqdm(total=llen + tmp_size, unit='B', unit_scale=True, colour='#31a8ff')
                progress_bar.update(tmp_size)
                with open(tmp_file, 'ab') as f:
                    for content in resp.iter_content(chunk_size=Download._DOWNLOAD_CHUNK_SIZE):
                        progress_bar.update(len(content))
                        f.write(content)
            os.renames(tmp_file, file_path)
        finally:
            if progress_bar:
                progress_bar.close()

        self._auth.log.info(f'文件下载完成 {file_path}')
        return file_path

    def download_files(self, files: List[BaseFile], local_folder: str = '.') -> List[str]:
        """批量下载文件

        :param files: BaseFile 对象列表, 或者包含 name 和  download_url 属性的对象
        :param local_folder: 目标文件夹, 表示要下载到哪个文件夹
        :return: 返回下载完成后的本地文件路径列表

        :Example:
        >>> from aligo import Aligo
        >>> ali = Aligo()
        >>> file_path = ali.download_files([ali.get_file_by_path('xxx.mp3')])
        >>> print(file_path)
        """
        rt = []
        for file in files:
            file_path = os.path.join(local_folder, file.name)
            try:
                file_path = self._core_download_file(file_path, file.download_url)
            except:
                file = self._core_get_file(GetFileRequest(file_id=file.file_id, drive_id=file.drive_id))
                file_path = self._core_download_file(file_path, file.download_url)
            rt.append(file_path)
        return rt



"""..."""
from dataclasses import dataclass

from .DataClass import DataClass


@dataclass
class VideoPreviewSprite(DataClass):
    """..."""
    col: int = None
    count: int = None
    frame_count: int = None
    frame_height: int = None
    frame_width: int = None
    row: int = None
    status: str = None



"""drive"""

from aligo.core import *
from aligo.request import *
from aligo.types import *


class Drive(Core):
    """..."""

    def get_drive(self, drive_id: str = None) -> BaseDrive:
        """
        获取 drive 信息
        :param drive_id: Optional[str] drive id，默认：None，如果为 None，则返回当前登录用户的 drive 信息
        :return: [BaseDrive]

        用法示例：
        >>> from aligo import Aligo
        >>> ali = Aligo()
        >>> result = ali.get_drive()
        >>> print(result)
        """
        body = GetDriveRequest(drive_id=drive_id)
        return self._core_get_drive(body)



"""..."""
from dataclasses import dataclass, field
from typing import Dict, List

from .DataClass import DataClass
from .Enum import *
from .ImageMedia import ImageMedia
from .VideoMedia import VideoMedia
from .VideoPreview import VideoPreview
from .FieldsInfo import FieldsInfo


@dataclass
class BaseFile(DataClass):
    """..."""
    name: str = None
    type: BaseFileType = None
    file_id: str = None
    parent_file_id: str = None
    category: BaseFileCategory = field(default=None, repr=False)
    size: int = field(default=None, repr=False)
    created_at: str = field(default=None, repr=False)
    content_type: str = field(default=None, repr=False)
    description: str = field(default=None, repr=False)
    content_hash: str = field(default=None, repr=False)
    content_hash_name: BaseFileContentHashName = field(default=None, repr=False)
    crc64_hash: str = field(default=None, repr=False)
    domain_id: str = field(default=None, repr=False)
    download_url: str = field(default=None, repr=False)
    drive_id: str = field(default=None, repr=False)
    encrypt_mode: str = field(default=None, repr=False)
    file_extension: str = field(default=None, repr=False)
    hidden: bool = field(default=None, repr=False)
    image_media_metadata: ImageMedia = field(default=None, repr=False)
    labels: list = field(default=None, repr=False)
    meta: str = field(default=None, repr=False)
    mime_extension: str = field(default=None, repr=False)
    mime_type: str = field(default=None, repr=False)
    punish_flag: int = field(default=None, repr=False)
    starred: bool = field(default=None, repr=False)
    status: str = field(default=None, repr=False)
    streams_url_info: Dict = field(default=None, repr=False)
    streams_info: Dict = field(default=None, repr=False)
    thumbnail: str = field(default=None, repr=False)
    trashed: bool = field(default=None, repr=False)
    trashed_at: str = field(default=None, repr=False)
    updated_at: str = field(default=None, repr=False)
    upload_id: str = field(default=None, repr=False)
    url: str = field(default=None, repr=False)
    user_meta: str = field(default=None, repr=False)
    video_media_metadata: VideoMedia = field(default=None, repr=False)
    video_preview_metadata: VideoPreview = field(default=None, repr=False)
    location: str = field(default=None, repr=False)
    action_list: List[str] = field(default_factory=list, repr=False)
    user_tags: Dict = field(default=None, repr=False)
    last_modifier_type: str = field(default=None, repr=False)
    last_modifier_id: str = field(default=None, repr=False)
    last_modifier_name: str = field(default=None, repr=False)
    creator_type: str = field(default=None, repr=False)
    creator_id: str = field(default=None, repr=False)
    creator_name: str = field(default=None, repr=False)
    revision_id: str = field(default=None, repr=False)
    sync_flag: bool = field(default=None, repr=False)
    sync_device_flag: bool = field(default=None, repr=False)
    sync_meta: str = field(default=None, repr=False)
    ex_fields_info: FieldsInfo = field(default=None, repr=False)



"""..."""

from aligo.core import *
from aligo.core.Config import *
from aligo.request import *
from aligo.types import *


class Drive(BaseAligo):
    """..."""

    def _core_get_drive(self, body: GetDriveRequest) -> BaseDrive:
        """..."""
        response = self._post(V2_DRIVE_GET, body=body)
        return self._result(response, BaseDrive)

    def get_default_drive(self) -> BaseDrive:
        """
        Get default drive.
        :return: [BaseDrive]

        :Example:
        >>> from aligo import Aligo
        >>> ali = Aligo()
        >>> drive = ali.drive.get_default_drive()
        >>> print(drive.name)
        """
        if self._default_drive is None:
            response = self._post(V2_DRIVE_GET_DEFAULT_DRIVE, body=GetDefaultDriveRequest(self._auth.token.user_id))
            self._default_drive = self._result(response, BaseDrive)
        return self._default_drive



"""..."""

from dataclasses import dataclass, field

from .DataClass import DataClass


@dataclass
class BaseDrive(DataClass):
    """..."""
    drive_id: str = None
    used_size: int = None
    total_size: int = None
    drive_name: str = field(default=None, repr=False)
    owner: str = field(default=None, repr=False)
    description: str = field(default=None, repr=False)
    drive_type: str = field(default=None, repr=False)
    creator: str = field(default=None, repr=False)
    domain_id: str = field(default=None, repr=False)
    status: str = field(default=None, repr=False)
    store_id: str = field(default=None, repr=False)
    owner_type: str = field(default=None, repr=False)
    relative_path: str = field(default=None, repr=False)
    encrypt_mode: str = field(default=None, repr=False)
    encrypt_data_access: bool = field(default=None, repr=False)
    permission: str = field(default=None, repr=False)
    created_at: str = field(default=None, repr=False)
    subdomain_id: str = field(default=None, repr=False)



"""..."""
import hashlib
import os
import shutil
from pathlib import Path
from typing import Callable
from typing import Dict
from typing import Optional

import arrow

from aligo.core import *
from aligo.request import GetFileListRequest, MoveFileToTrashRequest, CreateFolderRequest, BatchMoveToTrashRequest
from aligo.types import BaseFile
from .Download import Download


class SyncFolder(Core):
    """sync folder"""

    def sync_folder(
            self,
            local_folder: str,
            remote_folder: str,
            flag: Optional[bool] = None,
            follow_delete: bool = False,
            file_filter: Callable[[str], bool] = lambda x: False,
            ignore_content: bool = False,
            drive_id: str = None):
        """
        sync folder
        :param local_folder: 本地文件夹路径
        :param remote_folder: 云端文件夹 file_id
        :param flag: 同步标志（类型），默认：None, 另外可选：True, False
            - 冲突是指，程序无法自动处理的情况，例如：两端具有同名文件，但一个是文件，另一个是文件夹
            - 保持文件最新是指，两端具有同名文件，则下载或上传最新(根据最后修改时间)的文件，不受哪端为主的约束，三种模式处理方式均相同
            - 以某端为主
                - 会影响程序解决冲突的方式：删除次端冲突，保持主端文件。
                - 主端存在的，会同步到次端，反之则不会，并且会根据 `follow_delete`(是否跟随删除) 参数决定是否删除次端文件 
            None：
                - 不自动解决冲突
                - 不存在主次端
                - 同步所有文件，本地有的就上传，远端有的就下载，都有的看是否冲突，不冲突则保持最新
                - 此时会忽略 `follow_delete` 参数
            True：以本地为主
            False：以云端为主
        :param follow_delete: 次端是否跟随主端删除
        :param file_filter: 文件过滤函数，参数文件名，返回值为True则过滤，可用于实现 只同步 特定文件 或 排除某些文件
        :param ignore_content: 是否忽略文件内容，默认：False. 来自 issues/19 的功能请求
        :param drive_id: 云端文件夹 drive_id
        :return:

        :Example:
        >>> from aligo import Aligo
        >>> ali = Aligo()
        >>> ali.sync_folder('<本地路径>', '<云端文件夹 file_id>')
        """
        if flag is None:
            self._auth.log.info('sync_folder: 双端同步')
        elif flag:
            self._auth.log.info('sync_folder: 本地为主')
        else:
            self._auth.log.info('sync_folder: 云端为主')
        # 不管三七二十一，先创建一波
        Path(local_folder).mkdir(parents=True, exist_ok=True)
        self.__sync_folder(local_folder, remote_folder, flag, follow_delete, file_filter, ignore_content, drive_id)

    def __sync_folder(
            self,
            local_folder: str,
            remote_folder: str,
            flag: Optional[bool] = True,
            follow_delete: bool = False,
            file_filter: Callable[[str], bool] = lambda x: False,
            ignore_content: bool = False,
            drive_id: str = None):
        """..."""
        # 获取local_folder下的所有文件 # 转为 {文件名:文件路径} 字典模式
        local_files = {}
        for f in os.listdir(local_folder):
            local_file = os.path.join(local_folder, f)
            if file_filter(f):
                self._auth.log.debug(f'过滤本地文件 {local_file}')
                continue
            local_files[f] = local_file
        # 获取remote_folder下的所有文件 # 转为 {文件名:BaseFile对象} 字典模式
        remote_files: Dict[str, BaseFile] = {}
        for f in self._core_get_file_list(
                GetFileListRequest(remote_folder, drive_id=drive_id)
        ):
            remote_file = f.name
            if file_filter(remote_file):
                self._auth.log.debug(f'过滤云端文件 {remote_file}')
                continue
            remote_files[remote_file] = f

        if flag is None:
            self.__sync_all(drive_id, file_filter, flag, follow_delete, local_files,
                            local_folder, remote_files, remote_folder, ignore_content)
        elif flag:
            self.__sync_local(drive_id, follow_delete, local_files, remote_files,
                              remote_folder, flag, file_filter, local_folder, ignore_content)
        else:
            self.__sync_remote(drive_id, follow_delete, local_files, remote_files,
                               local_folder, flag, file_filter, ignore_content)

    def __sync_all(self, drive_id, file_filter, flag, follow_delete, local_files, local_folder, remote_files,
                   remote_folder, ignore_content):
        # 双端同步
        for f in local_files:
            local_file = local_files[f]
            # 文件夹则递归
            if os.path.isdir(local_file):
                # 判断云端是否存在
                if f in remote_files:
                    # 如果有，则判断是不是文件夹，不是就删除
                    remote_file = remote_files[f]
                    remote_files.pop(f)
                    if remote_file.type == 'file':
                        # 跳过冲突
                        self._auth.log.warning(f'冲突：本地是文件夹，云端是文件，不处理 {f}')
                        continue
                else:
                    # 如果没有，则创建文件夹
                    self._auth.log.debug(f'本地是文件夹，云端不存在，创建云端文件夹，并递归 {f}')
                    remote_file = self._core_create_folder(
                        CreateFolderRequest(name=f, parent_file_id=remote_folder, drive_id=drive_id,
                                            check_name_mode='overwrite')
                    )
                self.__sync_folder(local_file, remote_file.file_id, flag, follow_delete, file_filter, drive_id)
                continue

            # 如果本地文件存在，且在云端也存在，则依次比较 size，sha1，时间戳
            if f in remote_files:
                remote_file = remote_files[f]
                # 先判断文件类型
                if remote_file.type == 'folder':
                    # 没有主次端，不自动处理冲突
                    self._auth.log.warning(f'冲突：本地为文件，云端为文件夹，不处理 {f}')
                    remote_files.pop(f)
                    continue

                # 跳过对比文件内容
                if ignore_content:
                    self._auth.log.warning(f'忽略文件内容: 不处理 {f}')
                    remote_files.pop(f)
                    continue

                # 获取 f 文件大小
                local_size = os.path.getsize(local_file)
                # 比较大小
                if local_size == remote_file.size:
                    # 计算 f 文件 sha1
                    local_sha1 = self._core_sha1(local_file).lower()
                    # 如果sha1值相同，则跳过
                    if local_sha1 == remote_file.content_hash.lower():
                        self._auth.log.debug(f'文件 {f} 已存在，且sha1值相同，跳过')
                        remote_files.pop(f)
                        continue
                # 否则对比时间戳，删除较旧文件，同步最新文件
                # 获取 f 文件的最后修改时间
                local_time = os.path.getmtime(local_file)
                local_time = int(local_time)
                # 获取云端文件的最后修改时间
                remote_time = remote_file.updated_at
                # 将UTC时间字符串转换为时间戳
                remote_time = arrow.get(remote_time).timestamp()
                remote_time = int(remote_time)
                # 比较时间戳，删除较旧文件，同步最新文件
                if local_time > remote_time:
                    # 覆盖方式上传文件
                    self._auth.log.debug(f'本地文件较新，上传本地文件 {f}')
                    self.upload_file(file_path=local_file, parent_file_id=remote_folder, name=f,
                                     drive_id=drive_id, check_name_mode='overwrite')
                elif local_time < remote_time:
                    # 云端较新，删除本地，下载云端文件
                    self._auth.log.debug(f'云端较新，删除本地，下载云端文件 {f}')
                    os.remove(local_file)
                    self.download_files([remote_file], local_folder)
                # 在字典中删除remote_file
                remote_files.pop(f)
            else:
                # 不存在则直接上传
                self._auth.log.debug(f'云端不存在，上传本地文件 {f}')
                self.upload_file(file_path=local_file, parent_file_id=remote_folder, name=f,
                                 drive_id=drive_id, check_name_mode='overwrite')
        # 剩下的就是云端文件夹中有，本地文件夹中没有的文件，下载到本地
        if len(remote_files) != 0:
            self._auth.log.debug(f'本地不存在，下载云端文件 {len(remote_files)} {list(remote_files.keys())}')
            for remote_file in remote_files.values():
                if remote_file.type == 'file':
                    self.download_files([remote_file], local_folder)
                else:
                    Download.download_folder(self, remote_file.file_id, local_folder)

    def __sync_remote(self, drive_id, follow_delete, local_files, remote_files,
                      local_folder, flag, file_filter, ignore_content):
        # 以云端为主
        for f in remote_files:
            remote_file = remote_files[f]
            # 文件夹则递归
            if remote_file.type == 'folder':
                # 判断本地是否存在
                if f in local_files:
                    # 如果有，则判断是否是文件夹，不是就删除
                    local_file = local_files[f]
                    if os.path.isfile(local_file):
                        # 删除本地文件，创建本地文件夹
                        self._auth.log.debug(f'云端是文件夹，本地是文件，删除本地文件，并创建文件夹 {local_file}')
                        os.remove(local_file)
                        os.mkdir(local_file)
                    local_files.pop(f)
                else:
                    # 如果没有，则创建本地文件夹
                    local_file = os.path.join(local_folder, f)
                    self._auth.log.debug(f'云端是文件夹，本地没有，创建文件夹，并递归 {local_file}')
                    os.mkdir(local_file)
                self.__sync_folder(local_file, remote_file.file_id, flag, follow_delete, file_filter, drive_id)
                continue

            # 如果云端文件存在，且在本地也存在，则依次比较 size，sha1，时间戳
            if f in local_files:
                local_file = local_files[f]
                # 先判断文件类型
                if os.path.isdir(local_file):
                    # 云端为文件，本地却是文件夹，则递归删除文件夹，再下载文件
                    self._auth.log.debug('云端为文件，本地却是文件夹，则递归删除文件夹，再下载文件')
                    # 递归删除文件夹
                    shutil.rmtree(local_file)
                    # 下载文件
                    self.download_files([remote_file], local_folder)
                    local_files.pop(f)
                    continue

                # 跳过对比文件内容
                if ignore_content:
                    self._auth.log.warning(f'忽略文件内容: 不处理 {f}')
                    remote_files.pop(f)
                    continue

                # 获取 f 文件大小
                remote_size = remote_file.size
                # 比较大小
                if remote_size == os.path.getsize(local_file):
                    # 计算 f 文件 sha1
                    local_sha1 = self._core_sha1(local_file).lower()
                    # 如果sha1值相同，则跳过
                    if local_sha1 == remote_file.content_hash.lower():
                        self._auth.log.debug(f'文件 {local_file} 已存在，且sha1值相同，跳过')
                        local_files.pop(f)
                        continue
                # 否则对比时间戳，删除较旧文件，同步最新文件
                # 获取 f 文件的最后修改时间
                remote_time = remote_file.updated_at
                # 将UTC时间字符串转换为时间戳
                remote_time = arrow.get(remote_time).timestamp()
                remote_time = int(remote_time)
                # 获取本地文件的最后修改时间
                local_time = os.path.getmtime(local_file)
                local_time = int(local_time)
                # 比较时间戳，删除较旧文件，同步最新文件
                if local_time < remote_time:
                    # 先删除本地文件，再下载
                    self._auth.log.debug(f'云端文件较新，先删除本地文件，再下载云端文件 {local_file}')
                    os.remove(local_file)
                    self.download_files([remote_file], local_folder)
                elif local_time < remote_time:
                    # 云端较新，删除本地，下载云端文件
                    self._auth.log.debug(f'云端较新，删除本地，下载云端文件 {f}')
                    os.remove(local_file)
                    self.download_files([remote_file], local_folder)
                # 在字典中删除local_file
                local_files.pop(f)
            else:
                # 不存在直接下载
                local_file = os.path.join(local_folder, f)
                self._auth.log.debug(f'本地不存在，下载云端文件 {local_file}')
                self.download_files([remote_file], local_folder)

        # local_files剩下的就是多余的
        if follow_delete and len(local_files) != 0:
            self._auth.log.debug(f'删除本地文件 {len(local_files)} {list(local_files.values())}')
            for f in local_files:
                local_file = local_files[f]
                if os.path.isdir(local_file):
                    shutil.rmtree(local_file)
                else:
                    os.remove(local_file)

    def __sync_local(self, drive_id, follow_delete, local_files, remote_files,
                     remote_folder, flag, file_filter, local_folder, ignore_content):
        # 以本地为主
        for f in local_files:
            local_file = local_files[f]
            # 文件夹则递归
            if os.path.isdir(local_file):
                # 判断云端是否存在
                if f in remote_files:
                    # 如果有，则判断是不是文件夹，不是就删除
                    remote_file = remote_files[f]
                    if remote_file.type == 'file':
                        # 删除云端文件并创建文件夹
                        self._auth.log.debug(f'本地是文件夹，云端是文件，删除云端文件，并在云端创建文件夹 {f}')
                        self._core_move_file_to_trash(
                            MoveFileToTrashRequest(file_id=remote_file.file_id, drive_id=drive_id)
                        )
                        remote_file = self._core_create_folder(
                            CreateFolderRequest(name=f, parent_file_id=remote_folder, drive_id=drive_id,
                                                check_name_mode='overwrite')
                        )
                    remote_files.pop(f)
                else:
                    # 如果没有，则创建文件夹
                    self._auth.log.debug(f'本地是文件夹，云端不存在，创建云端文件夹，并递归 {f}')
                    remote_file = self._core_create_folder(
                        CreateFolderRequest(name=f, parent_file_id=remote_folder, drive_id=drive_id,
                                            check_name_mode='overwrite')
                    )
                self.__sync_folder(local_file, remote_file.file_id, flag, follow_delete, file_filter, drive_id)
                continue

            # 如果本地文件存在，且在云端也存在，则依次比较 size，sha1，时间戳
            if f in remote_files:
                remote_file = remote_files[f]
                # 先判断文件类型
                if remote_file.type == 'folder':
                    # 以本地文件夹为主，那只有删除云端文件夹，再上传本地文件
                    self._auth.log.debug(f'本地为文件，云端为文件夹，删除云端文件夹，并上传本地文件 {f}')
                    self._core_move_file_to_trash(
                        MoveFileToTrashRequest(file_id=remote_file.file_id, drive_id=drive_id)
                    )
                    self._auth.log.debug(f'上传本地文件 {f}')
                    self.upload_file(file_path=local_file, parent_file_id=remote_folder, name=f,
                                     drive_id=drive_id, check_name_mode='overwrite')
                    remote_files.pop(f)
                    continue

                # 跳过对比文件内容
                if ignore_content:
                    self._auth.log.warning(f'忽略文件内容: 不处理 {f}')
                    remote_files.pop(f)
                    continue

                # 获取 f 文件大小
                local_size = os.path.getsize(local_file)
                # 比较大小
                if local_size == remote_file.size:
                    # 计算 f 文件 sha1
                    local_sha1 = self._core_sha1(local_file).lower()
                    # 如果sha1值相同，则跳过
                    if local_sha1 == remote_file.content_hash.lower():
                        self._auth.log.debug(f'文件 {f} 已存在，且sha1值相同，跳过')
                        remote_files.pop(f)
                        continue
                # 否则对比时间戳，删除较旧文件，同步最新文件
                # 获取 f 文件的最后修改时间
                local_time = os.path.getmtime(local_file)
                local_time = int(local_time)
                # 获取云端文件的最后修改时间
                remote_time = remote_file.updated_at
                # 将UTC时间字符串转换为时间戳
                remote_time = arrow.get(remote_time).timestamp()
                remote_time = int(remote_time)
                # 比较时间戳，删除较旧文件，同步最新文件
                if local_time > remote_time:
                    # 覆盖方式上传文件
                    self._auth.log.debug(f'本地文件较新，上传本地文件 {f}')
                    self.upload_file(file_path=local_file, parent_file_id=remote_folder, name=f,
                                     drive_id=drive_id, check_name_mode='overwrite')
                elif local_time < remote_time:
                    # 云端较新，删除本地，下载云端文件
                    self._auth.log.debug(f'云端较新，删除本地，下载云端文件 {f}')
                    os.remove(local_file)
                    self.download_files([remote_file], local_folder)
                # 在字典中删除remote_file
                remote_files.pop(f)
            else:
                # 不存在则直接上传
                self._auth.log.debug(f'云端不存在，上传本地文件 {f}')
                self.upload_file(file_path=local_file, parent_file_id=remote_folder, name=f,
                                 drive_id=drive_id, check_name_mode='overwrite')
        # 判断是否删除云端文件
        if follow_delete and len(remote_files) != 0:
            self._auth.log.debug(f'删除云端文件 {len(remote_files)} {list(remote_files.keys())}')
            for _ in self._core_batch_move_to_trash(BatchMoveToTrashRequest(
                    drive_id=drive_id, file_id_list=[f.file_id for f in remote_files.values()]
            )):
                pass

    @staticmethod
    def _core_sha1(param):
        """计算文件sha1"""
        sha1 = hashlib.sha1()
        with open(param, 'rb') as f:
            while True:
                data = f.read(8192)
                if not data:
                    break
                sha1.update(data)
        return sha1.hexdigest()



"""..."""
from .Audio import Audio
from .Copy import Copy
from .Create import Create
from .Download import Download
from .Drive import Drive
from .File import File
from .Move import Move
from .Other import Other
from .Recyclebin import Recyclebin
from .Search import Search
from .Share import Share
from .Star import Star
from .User import User
from .Video import Video


class Core(
    Audio,
    Copy,
    Create,
    Drive,
    File,
    Move,
    Download,
    Recyclebin,
    Search,
    Share,
    User,
    Video,
    Star,
    Other,
):
    """..."""



"""Aligo class"""
import json
import logging
from typing import Callable, Dict, Tuple

from typing_extensions import NoReturn

from .Audio import Audio, aligo_config_folder
from .Copy import Copy
from .Create import Create
from .CustomShare import CustomShare
from .Download import Download
from .Drive import Drive
from .File import File
from .Move import Move
from .Other import Other
from .Recyclebin import Recyclebin
from .Search import Search
from .Share import Share
from .Star import Star
from .SyncFolder import SyncFolder
from .Update import Update
from .Video import Video


class Aligo(
    Audio,
    Video,
    Copy,
    Create,
    Drive,
    File,
    Move,
    Download,
    Recyclebin,
    Search,
    Share,
    CustomShare,
    Star,
    Update,
    Other,
    SyncFolder,
):
    """阿里云盘"""

    def __init__(
            self,
            name: str = 'aligo',
            refresh_token: str = None,
            show: Callable[[str], NoReturn] = None,
            level: int = logging.DEBUG,
            use_aria2: bool = False,
            proxies: Dict = None,
            port: int = None,
            email: Tuple[str, str] = None,
    ):
        """
        Aligo
        :param name: (可选, 默认: aligo) 配置文件名称, 便于使用不同配置文件进行身份验证
        :param refresh_token:
        :param show: (可选) 显示二维码的函数
        :param level: (可选) 控制控制台输出
        :param use_aria2: [bool] 是否使用 aria2 下载
        :param proxies: (可选) 自定义代理 [proxies={"https":"localhost:10809"}],支持 http 和 socks5（具体参考requests库的用法）
        :param port: (可选) 开启 http server 端口，用于网页端扫码登录. 提供此值时，将不再弹出或打印二维码
        :param email: (可选) 发送扫码登录邮件 ("接收邮件的邮箱地址", "防伪字符串"). 提供此值时，将不再弹出或打印二维码
            关于防伪字符串: 为了方便大家使用, aligo 自带公开邮箱, 省去邮箱配置的麻烦.
                        所以收到登录邮件后, 一定要对比确认防伪字符串和你设置一致才可扫码登录, 否则将导致: 包括但不限于云盘文件泄露.
            关于防伪字符串: 为了方便大家使用, aligo 自带公开邮箱, 省去邮箱配置的麻烦.
                        所以收到登录邮件后, 一定要对比确认防伪字符串和你设置一致才可扫码登录, 否则将导致: 包括但不限于云盘文件泄露.
            关于防伪字符串: 为了方便大家使用, aligo 自带公开邮箱, 省去邮箱配置的麻烦.
                        所以收到登录邮件后, 一定要对比确认防伪字符串和你设置一致才可扫码登录, 否则将导致: 包括但不限于云盘文件泄露.

        level, use_aria2, proxies, port, email 可以通过 配置文件 配置默认值，在 <用户家目录>/.aligo/config.json5 中
        ```json5
        {
          "level": 10,
          "use_aria2": false,
          "proxies": {
            "https": "http://localhost:10809",
            // "https": "socks5://localhost:10808", # 不支持注释，写的时候删掉
          },
          "port": 8080,
          "email": ["邮箱地址", "防伪字符串"]
        }
        ```
        """
        config = aligo_config_folder / 'config.json5'
        if config.exists():
            config = json.loads(config.open().read())
            if level == logging.DEBUG and 'level' in config:
                level = config.get('level')
            if not use_aria2:
                use_aria2 = config.get('use_aria2')
            if proxies is None:
                proxies = config.get('proxies')
            if port is None:
                port = config.get('port')
            if email is None:
                email = config.get('email')

        super().__init__(
            name,
            refresh_token,
            show,
            level,
            use_aria2,
            proxies,
            port,
            email,
        )



"""..."""
# 导包基本原则
# 1. 包内相对导入: from .DataClass import DataClass
# 2. 包外包导入: from aligo.dataobj import xxx
from dataclasses import dataclass, field
from typing import List

from .Enum import VideoTemplateID
from .DataClass import DataClass


@dataclass
class LiveTranscodingMeta(DataClass):
    """..."""
    ts_segment: int = None
    ts_total_count: int = None
    ts_pre_count: int = None


@dataclass
class Meta(DataClass):
    """..."""
    duration: float = None
    width: int = None
    height: int = None
    live_transcoding_meta: LiveTranscodingMeta = None


@dataclass
class LiveTranscodingTask(DataClass):
    """..."""
    template_id: VideoTemplateID = None
    status: str = None
    stage: str = None
    url: str = None


@dataclass
class VideoPreviewPlayInfo(DataClass):
    """..."""
    category: str = None
    meta: Meta = None
    live_transcoding_task_list: List[LiveTranscodingTask] = field(default_factory=list)



"""Other"""

from aligo.request import GetFilePathRequest
from aligo.response import GetFilePathResponse
from .BaseAligo import BaseAligo
from .Config import *


class Other(BaseAligo):
    """Other"""

    def _core_get_path(self, body: GetFilePathRequest) -> GetFilePathResponse:
        """get_path 获取当前文件的路径, 父级目录"""
        response = self._post(ADRIVE_V1_FILE_GET_PATH, body=body)
        return self._result(response, GetFilePathResponse)



"""..."""
from dataclasses import dataclass

from .DataClass import DataClass


@dataclass
class CroppingBoundary(DataClass):
    """..."""
    height: int = None
    left: int = None
    top: int = None
    width: int = None



"""..."""
from dataclasses import dataclass

from aligo.types import DataClass, VideoPreviewPlayInfo


@dataclass
class GetVideoPreviewPlayInfoResponse(DataClass):
    """..."""
    domain_id: str = None
    drive_id: str = None
    file_id: str = None
    video_preview_play_info: VideoPreviewPlayInfo = None



"""..."""
# 主机
API_HOST = 'https://api.aliyundrive.com'
AUTH_HOST = 'https://auth.aliyundrive.com'
WEBSV_HOST = 'https://websv.aliyundrive.com'
PASSPORT_HOST = 'https://passport.aliyundrive.com'
MEMBER_HOST = 'https://member.aliyundrive.com'

# 路径
NEWLOGIN_LOGIN_DO = '/newlogin/login.do'
NEWLOGIN_SMS_SEND_DO = '/newlogin/sms/send.do'
NEWLOGIN_SMS_LOGIN_DO = '/newlogin/sms/login.do'
NEWLOGIN_QRCODE_QUERY_DO = '/newlogin/qrcode/query.do'
NEWLOGIN_QRCODE_GENERATE_DO = '/newlogin/qrcode/generate.do'

V2_OAUTH_AUTHORIZE = '/v2/oauth/authorize'
V2_OAUTH_TOKEN_LOGIN = '/v2/oauth/token_login'

TOKEN_GET = '/token/get'
TOKEN_REFRESH = '/token/refresh'  # TODO: 这个接口可能是旧的接口, 待删除
V2_ACCOUNT_TOKEN = '/v2/account/token'

V2_USER_GET = '/v2/user/get'

V2_DRIVE_GET = '/v2/drive/get'
V2_DRIVE_GET_DEFAULT_DRIVE = '/v2/drive/get_default_drive'

V2_FILE_GET = '/v2/file/get'
V2_FILE_LIST = '/v2/file/list'
ADRIVE_V3_FILE_LIST = '/adrive/v3/file/list'
V3_FILE_UPDATE = '/v3/file/update'
V2_FILE_MOVE = '/v2/file/move'
V2_FILE_COPY = '/v2/file/copy'
V2_FILE_SEARCH = '/v2/file/search'
V2_FILE_CREATE = '/v2/file/create'
V2_FILE_COMPLETE = '/v2/file/complete'
ADRIVE_V2_FILE_CREATE = '/adrive/v2/file/create'
V2_FILE_CREATE_WITH_PROOF = '/v2/file/create_with_proof'
ADRIVE_V2_FILE_CREATEWITHFOLDERS = '/adrive/v2/file/createWithFolders'
V2_FILE_GET_UPLOAD_URL = '/v2/file/get_upload_url'

V2_AIMS_SEARCH = '/v2/aims/search'

V2_FILE_GET_DOWNLOAD_URL = '/v2/file/get_download_url'

V2_RECYCLEBIN_TRASH = '/v2/recyclebin/trash'
V2_RECYCLEBIN_LIST = '/v2/recyclebin/list'
V2_RECYCLEBIN_RESTORE = '/v2/recyclebin/restore'

V2_DATABOX_GET_PERSONAL_INFO = '/v2/databox/get_personal_info'
V2_DATABOX_GET_AUDIO_PLAY_INFO = '/v2/databox/get_audio_play_info'
V2_DATABOX_GET_VIDEO_PLAY_INFO = '/v2/databox/get_video_play_info'

V2_FILE_LIST_BY_CUSTOM_INDEX_KEY = '/v2/file/list_by_custom_index_key'

V2_SHARE_LINK_UPDATE = '/v2/share_link/update'
ADRIVE_V3_SHARE_LINK_LIST = '/adrive/v3/share_link/list'
ADRIVE_V2_SHARE_LINK_CREATE = '/adrive/v2/share_link/create'
ADRIVE_V2_SHARE_LINK_CANCEL = '/adrive/v2/share_link/cancel'
ADRIVE_V2_SHARE_LINK_GET_SHARE_BY_ANONYMOUS = '/adrive/v2/share_link/get_share_by_anonymous'
V2_SHARE_LINK_GET_SHARE_TOKEN = '/v2/share_link/get_share_token'
V2_FILE_GET_SHARE_LINK_DOWNLOAD_URL = '/v2/file/get_share_link_download_url'

V3_BATCH = '/v3/batch'
ADRIVE_V2_BATCH = '/adrive/v2/batch'

V1_USERS_REWARDS = '/v1/users/rewards'

ADRIVE_V1_FILE_GET_PATH = '/adrive/v1/file/get_path'

# m3u8
V2_FILE_GET_VIDEO_PREVIEW_PLAY_INFO = '/v2/file/get_video_preview_play_info'

# 参数
CLIENT_ID = '25dzX3vbYqktVxyX'
UNI_PARAMS = {'appName': 'aliyun_drive'}
UNI_HEADERS = {
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.114 Safari/537.36',
    # 'User-Agent': 'AliApp(yunpan/2.3.0) com.alicloud.databox/15287370 Channel/263200@rimet_android_2.3.0 language/zh-CN /Android Mobile/OnePlus GM1900',
    # 'Referer': 'https://aliyundrive.com'
}
